// 対応点探索に関する関数群
#include <cv.h>
#include "correspond.h"
#include "fundamental.h"
#include <stdio.h>

#define DEF_TEMP_WIDTH 7
#define DEF_TEMP_HEIGHT 8

static int Width = DEF_TEMP_WIDTH;
static int Height = DEF_TEMP_HEIGHT;

/* ----- prototype ----- */
static double cvImageDistance(IplImage *img1,CvPoint2D32f point1, 
			      IplImage *img2,CvPoint2D32f point2);
// 画像間のL2ノルムを計算

static double cvImageCorrelation(IplImage *img1,CvPoint2D32f point1, 
				 IplImage *img2,CvPoint2D32f point2);
// 画像間の正規化相関を計算

static CvMat *makeScoreMap(IplImage *img1,IplImage *img2,
			   CvPoint2D32f *point1,CvPoint2D32f *point2,
			   int pnum1,int pnum2,
			   CvMat *F,
			   double img_thre);
// 2枚の画像の特徴点群について，点同士の相関マップを計算する


static CvMat *makeMaskAll1(int row,int col);
// 全ての要素が1のマスク画像を生成

static CvMyCorrespond *makeCorrespondFromSeq(CvSeq *seq,int num);
// CvSeq を対応点組に変換

/* ----- ----- ----- */


int searchCorrespond2(IplImage *img1,CvPoint2D32f *point1,int pnum1,
		      IplImage *img2,CvPoint2D32f *point2,int pnum2,
		      CvMyCorrespond **correspond_ret,int *cor_num_ret,
		      CvMat *F,
		      double cor_min,double length){
  // 2枚の画像間の組み合わせを計算する．
  // img1,img2 : 入力画像
  // point1,point2 : 特徴点，数が同じ必用はない
  // pnum1,pnum2 : 各特徴点の数
  // correspond_ret : 対応点組を格納，NULLであれば何も返さない
  // cor_num_ret : 対応点組の数
  // F : 計算過程で得られるF行列を格納
  // cor_min : 対応点相関の最小値
  // length : エピポーラ線との距離の閾値
  CvMat *status;
  CvMat *score,*mask;
  CvMyCorrespond *correspond;
  int cornum;

  // 点候補を一時格納しておくストレージ
  CvMemStorage *storage = cvCreateMemStorage(0);
  CvSeq *seq = cvCreateSeq(0,sizeof(CvSeq),sizeof(CvPoint2D32f),storage);

  int i;
  int rnum = 0;
  
  // 点同士の画像相関を計算
  score = makeScoreMap(img1,img2,point1,point2,pnum1,pnum2,NULL,length);
  mask = makeMaskAll1(pnum1,pnum2);
  
  // 相関の高い点組から順次対応点の候補とする
  while(1){
    double min,max;
    CvPoint minp,maxp;
    cvMinMaxLoc(score,&min,&max,&minp,&maxp,mask);
    if (max>cor_min){
      for (i=0;i<pnum1;i++)
	cvSetReal2D(mask,i,maxp.x,0);
      for (i=0;i<pnum2;i++)
	cvSetReal2D(mask,maxp.y,i,0);
      cvSeqPush(seq,point1+maxp.y);
      cvSeqPush(seq,point2+maxp.x);
    }else{
      break;
    }
  }
  
  cvReleaseMat(&mask);
  cvReleaseMat(&score);
  
  // 対応点組を作成
  correspond = makeCorrespondFromSeq(seq,2);
  cornum = seq->total/2;
  
  // F行列計算を利用して適切と思われる点組のみを抽出
  status = cvCreateMat(cornum,1,CV_32SC1);
  findFundamentalMat(correspond,cornum,F,length,10000,status);
  
  // inlier と判定された点を対応点として登録
  rnum = 0;
  for (i=0;i<cornum;i++)
    if (cvGetReal1D(status,i)!=0)
      rnum ++;
  if (correspond_ret!=NULL){
    *correspond_ret = (CvMyCorrespond*)calloc(rnum,sizeof(CvMyCorrespond));
    rnum = 0;
    for (i=0;i<cornum;i++)
      if (cvGetReal1D(status,i)){
	(*correspond_ret)[rnum]=correspond[i];
	rnum++;
      }
  }
  if (cor_num_ret!=NULL)
    *cor_num_ret = rnum;
  
  // 解放
  free(correspond);
  cvReleaseMat(&status);
  cvReleaseMemStorage(&storage);
  
  return 0;
}

int searchCorrespond3(IplImage *img1,CvPoint2D32f *point1,int pnum1,
		      IplImage *img2,CvPoint2D32f *point2,int pnum2,
		      IplImage *img3,CvPoint2D32f *point3,int pnum3,
		      CvMat *F12,CvMat *F13,CvMat *F23,
		      CvMyCorrespond **correspond_ret,
		      int *cor_num_ret,
		      double cor_min,double length){
  // 3枚の画像間の対応点を計算する
  // F行列を利用して正確さを向上させる
  // Fxy : x -> y の F 行列，NULLの場合は使用されない
  // それ以外は searchCorrespond2と同様

  double max3;
  int minnum1,minnum2,minnum3;
  CvMat *score12,*score13,*score23,*mask;
  int *used3;
  int i;
  
  // 点を一時格納しておくストレージ
  CvMemStorage *storage = cvCreateMemStorage(0);
  CvSeq *seq = cvCreateSeq(0,sizeof(CvSeq),sizeof(CvPoint2D32f),storage);

  // 各画像間の相関マップを作成
  score12 = makeScoreMap(img1,img2,point1,point2,pnum1,pnum2,F12,length);
  score13 = makeScoreMap(img1,img3,point1,point3,pnum1,pnum3,F13,length);
  score23 = makeScoreMap(img2,img3,point2,point3,pnum2,pnum3,F23,length);

  mask = makeMaskAll1(pnum1,pnum2);
  
  // 1 -> 2 の評価値を基準に対応点組を作成
  // 3の対応点に重複がでないように管理するフラグ
  used3 = (int*)calloc(pnum3,sizeof(int));
  while(1){
    double min,max;
    CvPoint minp,maxp;
    // 1 -> 2 の最小値を探索
    cvMinMaxLoc(score12,&min,&max,&minp,&maxp,mask);
    minnum1 = maxp.y;
    minnum2 = maxp.x;
    // 1->2 の最大相関が cor_min 以下であれば終了
    if (max<cor_min)
      break;
    
    // 3点目を探索(1->3 + 2->3 を評価値とする)
    // 1,2画像についてminnum1,minnum2点は使用済みとする
    for (i=0;i<pnum1;i++)
      cvSetReal2D(mask,i,minnum2,0);
    for (i=0;i<pnum2;i++)
      cvSetReal2D(mask,minnum1,i,0);
    max3 = cor_min*2;
    minnum3 = -1;

    // 1->3, 2->3 の相関が閾値以上となり，和が最大となる点を探索
    for (i=0;i<pnum3;i++){
      if (used3[i]==0){
	double s1 = cvmGet(score13,minnum1,i);
	double s2 = cvmGet(score23,minnum2,i);
	if (s1 > cor_min && s2 > cor_min && s1 + s2 > max){
	  minnum3 = i;
	  max3 = s1 + s2;
	}
      }
    }
    
    if (minnum3<0)
      continue;

    used3[minnum3] = 1;
    cvSeqPush(seq,&point1[minnum1]);
    cvSeqPush(seq,&point2[minnum2]);
    cvSeqPush(seq,&point3[minnum3]);
  }

  if (correspond_ret!=NULL)
    *correspond_ret = makeCorrespondFromSeq(seq,3);
  if (cor_num_ret!=NULL)
    *cor_num_ret = seq->total/3;
  

  // 後始末
  cvReleaseMat(&score12);
  cvReleaseMat(&score13);
  cvReleaseMat(&score23);
  cvReleaseMat(&mask);
  cvReleaseMemStorage(&storage);

  free(used3);


  return 0;
}

static double cvImageDistance(IplImage *img1,CvPoint2D32f point1, 
			      IplImage *img2,CvPoint2D32f point2){
  // 画像間のL2ノルムを計算
  // 画像をいちいち確保するので少々遅い
  IplImage *tmp1,*tmp2,*diff;
  CvScalar dis;
  CvRect roi;
  
  if (point1.x-Width<0 || point1.x+Width > img1->width||
      point1.y-Height<0 || point1.y+Height > img1->height||
      point2.x-Width<0 || point2.x+Width > img2->height||
      point2.y-Height<0 || point2.y+Height > img2->height)
    return -1;
  
  // 画像1から点の周辺画像を切り出し
  tmp1 = cvCreateImage(cvSize(Width * 2, Height * 2),
		       img1->depth,img1->nChannels);
  roi = cvGetImageROI(img1);
  cvSetImageROI(img1,cvRect(point1.x-Width,point1.y-Height,
			    Width * 2, Height * 2));
  cvCopyImage(img1,tmp1);
  cvSetImageROI(img1,roi);
  
  // 画像2から点の周辺画像を切り出し
  tmp2 = cvCreateImage(cvSize(Width * 2, Height * 2),
		       img2->depth,img2->nChannels);
  roi = cvGetImageROI(img2);
  cvSetImageROI(img2,cvRect(point2.x-Width,point2.y-Height,
			    Width * 2, Height * 2));
  cvCopyImage(img2,tmp2);
  cvSetImageROI(img2,roi);
  
  // 差分画像を確保
  diff = cvCreateImage(cvSize(Width * 2, Height * 2),
		       img1->depth,img1->nChannels);
  cvAbsDiff(tmp1,tmp2,diff);
  dis = cvAvg(diff);

  cvReleaseImage(&tmp1);
  cvReleaseImage(&tmp2);
  cvReleaseImage(&diff);

  return dis.val[0];
}

static double cvImageCorrelation(IplImage *img1,CvPoint2D32f point1, 
				 IplImage *img2,CvPoint2D32f point2){
  
  // 画像間の正規化相関を計算
  double ip,norm1,norm2;
  IplImage *tmp1,*tmp2;
  CvRect roi;

  if (point1.x-Width<0 || point1.x+Width > img1->width||
      point1.y-Height<0 || point1.y+Height > img1->height||
      point2.x-Width<0 || point2.x+Width > img2->width||
      point2.y-Height<0 || point2.y+Height > img2->height){
    return -1;
  }
  
  // 画像1から点の周辺画像を切り出し
  tmp1 = cvCreateImage(cvSize(Width * 2, Height * 2),
		       img1->depth,img1->nChannels);
  roi = cvGetImageROI(img1);
  cvSetImageROI(img1,cvRect(point1.x-Width,point1.y-Height,
			    Width * 2, Height * 2));
  cvCopyImage(img1,tmp1);
  cvSetImageROI(img1,roi);
  
  // 画像2から点の周辺画像を切り出し
  tmp2 = cvCreateImage(cvSize(Width * 2, Height * 2),
		       img2->depth,img2->nChannels);
  roi = cvGetImageROI(img2);
  cvSetImageROI(img2,cvRect(point2.x-Width,point2.y-Height,
			    Width * 2, Height * 2));
  cvCopyImage(img2,tmp2);
  cvSetImageROI(img2,roi);
  
  norm1 = cvDotProduct(tmp1,tmp1);
  norm2 = cvDotProduct(tmp2,tmp2);
  ip = cvDotProduct(tmp1,tmp2);

  cvReleaseImage(&tmp1);
  cvReleaseImage(&tmp2);

  cvResetImageROI(img1);
  cvResetImageROI(img2);

  return ip/sqrt(norm1*norm2);
}

static CvMat *makeScoreMap(IplImage *img1,IplImage *img2,
			   CvPoint2D32f *point1,CvPoint2D32f *point2,
			   int pnum1,int pnum2,
			   CvMat *F,
			   double img_thre){
  // 2枚の画像の特徴点群について，点同士の相関マップを計算する
  
  CvMat *score;
  int i,j;
  
  score = cvCreateMat(pnum1,pnum2,CV_64F);
  for (i=0;i<pnum1;i++)
    for (j=0;j<pnum2;j++)
      if (cvFPointDis(F,point1[i],point2[j])<img_thre){
	double dis = cvImageCorrelation(img1,point1[i],img2,point2[j]);
	cvmSet(score,i,j,dis);
      }else{
	cvmSet(score,i,j,-2);
      }
  return score;
}

static CvMat *makeMaskAll1(int row,int col){
  // 全ての要素が1のマスク画像を生成
  CvMat *mask = cvCreateMat(row,col,CV_8S);
  int i,j;

  for (i=0;i<row;i++)
    for (j=0;j<col;j++)
      cvSetReal2D(mask,i,j,1);
  return mask;
}

static CvMyCorrespond *makeCorrespondFromSeq(CvSeq *seq,int num){
  // CvSeq を対応点組に変換
  int i,j;
  CvMyCorrespond *correspond;
  CvSeqReader reader;
  
  if (seq->total%num!=0)
    return NULL;
  
  correspond = (CvMyCorrespond*)calloc(seq->total/num,sizeof(CvMyCorrespond));

  cvStartReadSeq(seq,&reader,0);
  for (i=0;i<seq->total/num;i++){
    correspond[i].num = num;
    for (j=0;j<num;j++){
      CvPoint2D32f *pt = (CvPoint2D32f*)(reader.ptr);
      correspond[i].point[j] = *pt;
      CV_NEXT_SEQ_ELEM(sizeof(CvPoint2D32f),reader);
    }
  }
  return correspond;
}


