// F行列周りの関数群
#include <cv.h>
#include "fundamental.h"

static int random_swap(int num,int swapnum,int *list){
  // list の中身をランダムに入れ替え
  int i,r;
  int swap;
  static CvRNG rngstate = cvRNG(0xffffffff);

  for (i=0;i<num;i++){
    r = (int)(cvRandInt(&rngstate) / (double) 0xffffffff * num);
    swap = list[i];
    list[i] = list[r];
    list[r] = swap;
  }
  
  return 0;
}

int drawEpiline(IplImage *img,CvPoint2D32f pt,CvPoint2D32f pt2,
		CvMat *F){
  // 画像にエピポーラ線を描画
  double line[3];
  CvPoint x1,x2;
  int i;
  double norm;

  for (i=0;i<3;i++)
    line[i] = pt.x * cvmGet(F,0,i) + pt.y * cvmGet(F,1,i) + cvmGet(F,2,i);
  norm = sqrt(line[0]*line[0]+line[1]*line[1]);
  line[0] /= norm;
  line[1] /= norm;
  line[2] /= norm;
  x1.x = 0; 
  x2.x = img->width-1;
  
  x1.y = -(line[2] + line[0] * x1.x)/line[1];
  x2.y = -(line[2] + line[0] * x2.x)/line[1];

  cvLine(img,x1,x2,cvScalar(255,0,0));
  
  return 0;
}

double cvFPointDis(CvMat *F,CvPoint2D32f pt1,CvPoint2D32f pt2){
  // F を使用してpt1, pt2 が対応点であるかどうかチェック
  double line[3];
  if (F==NULL) return 0;
  
  line[0] = 
    pt1.x * cvmGet(F,0,0) + pt1.y * cvmGet(F,0,1) + cvmGet(F,0,2);
  line[1] = 
    pt1.x * cvmGet(F,1,0) + pt1.y * cvmGet(F,1,1) + cvmGet(F,1,2);
  line[2] = 
    pt1.x * cvmGet(F,2,0) + pt1.y * cvmGet(F,2,1) + cvmGet(F,2,2);
  return 
    fabs((line[0] * pt2.x + line[1] * pt2.y + line[2])/
	 sqrt(line[0]*line[0] + line[1]*line[1]));
}

int cvFTest(CvMat *mat1,CvMat *mat2,
	    CvMat *F,double length,CvMat *status){
  // 与えられた点群に対して，Fのサポート数をチェック
  int i;
  int sup = 0;
  double dis;

  for (i=0;i<mat1->rows;i++){
    dis = cvFPointDis(F,
		      cvPoint2D32f(cvmGet(mat1,i,0),cvmGet(mat1,i,1)),
		      cvPoint2D32f(cvmGet(mat2,i,0),cvmGet(mat2,i,1)));
    if (fabs(dis)<length){
      cvSetReal1D(status,i,1);
      sup++;
    }else{
      cvSetReal1D(status,i,0);
    }
  }

  return sup;
}

int cvMyFindFundamentalMat(CvMat *mat1,CvMat *mat2,CvMat *F,
			   double val1, double val2,
			   int loop_num,
			   CvMat *status){
  // Fのロバスト性がいまいちなので，自前でRANSACを実装
  CvMat *p1,*p2,*Ftmp,*Fmax;
  int sup,max = 0 ;
  int loop;
  int *order;
  int num;
  int i;

  // RANSACに使用する点選択の準備
  order = (int*)calloc(mat1->rows,sizeof(int));
  for (i=0;i<mat1->rows;i++)
    order[i] = i;
  
  p1 = cvCreateMat(8,2,CV_64F);
  p2 = cvCreateMat(8,2,CV_64F);
  Ftmp = cvCreateMat(3,3,CV_64F);
  Fmax = cvCreateMat(3,3,CV_64F);

  for (loop=0;loop<loop_num;loop++){
    // 使用する点をランダムに選択
    random_swap(mat1->rows,8,order);
    for (int j=0;j<8;j++){
      cvmSet(p1,j,0,cvmGet(mat1,order[j],0));
      cvmSet(p1,j,1,cvmGet(mat1,order[j],1));
      cvmSet(p2,j,0,cvmGet(mat2,order[j],0));
      cvmSet(p2,j,1,cvmGet(mat2,order[j],1));
    }

    // 8点法を利用してF計算
    if (cvFindFundamentalMat(p1,p2,Ftmp,CV_FM_8POINT,1,val2)!=0){
      // 検証(F計算に使用する点が入るので必ず8以上のサポート)
      sup = cvFTest(mat1,mat2,Ftmp,val2,status);
      if (max<sup){
	max = sup;
	cvCopy(Ftmp,Fmax);
	sup = cvFTest(mat1,mat2,Fmax,val2,status);
      }
    }
  }
  cvReleaseMat(&p1);
  cvReleaseMat(&p2);
  
  cvCopy(Fmax,F);
  // inlierとなった全ての点を利用してFを再計算
  sup = cvFTest(mat1,mat2,F,val2,status);
  p1 = cvCreateMat(sup,2,CV_64F);
  p2 = cvCreateMat(sup,2,CV_64F);
  num = 0;
  for (i=0;i<mat1->rows;i++)
    if (cvGetReal1D(status,i)!=0){
      cvmSet(p1,num,0,cvmGet(mat1,i,0));
      cvmSet(p1,num,1,cvmGet(mat1,i,1));
      cvmSet(p2,num,0,cvmGet(mat2,i,0));
      cvmSet(p2,num,1,cvmGet(mat2,i,1));
      num++;
    }
  cvFindFundamentalMat(p1,p2,F,CV_FM_8POINT,1,val2);
  
  // 後始末
  cvReleaseMat(&p1);
  cvReleaseMat(&p2);
  cvReleaseMat(&Ftmp);
  cvReleaseMat(&Fmax);
  free(order);
  
  return 0;

}

static int cvmSetPointRow(CvMat *mat,int row,CvPoint2D32f pt){
  if (mat==NULL || mat->rows<=row) return -1;
  cvmSet(mat,row,0,pt.x);
  cvmSet(mat,row,1,pt.y);

  if (mat->cols==3)
    cvmSet(mat,row,2,1);
  
  return 0;
}

int findFundamentalMat(CvMyCorrespond *correspond,int num,
		       CvMat *F,double length,
		       int loopnum,CvMat *status){
  // 対応点組から F 行列を計算
  CvMat *mat[2];
  int i,j;

  mat[0] = cvCreateMat(num,2,CV_64F);
  mat[1] = cvCreateMat(num,2,CV_64F);

  // mat に格納
  for (i=0;i<num;i++)
    for (j=0;j<2;j++)
      cvmSetPointRow(mat[j],i,correspond[i].point[j]);

  // F 計算
  cvMyFindFundamentalMat(mat[0],mat[1],F,1,length,loopnum,status);

  cvReleaseMat(&mat[0]);
  cvReleaseMat(&mat[1]);

  return 0;
}
