/* よく使う線形演算 */
#include <cv.h>
#include <stdio.h>

#include "linear.h"

// SVDだとランク落ちに対応できないので固有値分解を利用
#define USE_EIGEN_VV

static int cvCreateMatEpsilonRecur(CvMatND *tensor,int dim,int nest,int *order);
static double cvGetEpsilonElement(int dim,int *oder);

static double sqr(double x){
  return x*x;
}

int cvGetMinEigenVector(CvMat *M,
			CvMat *ret_vector){
  CvMat *MtM;
  CvMat *evects,*evals;
  int ret=0;

  // MtM の行列確保
  MtM = cvCreateMat(M->cols,M->cols,CV_64F);

  // MtM を計算
  cvMulTransposed(M,MtM,1,NULL,1.0);
  
  // MtM を固有値分解
  evects = cvCreateMat(M->cols,M->cols,CV_64F);
  evals = cvCreateMat(M->cols,1,CV_64F);
#ifdef USE_EIGEN_VV
  // 固有値分解を利用する場合
  cvEigenVV(MtM,evects,evals,DBL_EPSILON);
#else
  // 特異値分解で代用する場合
  ret = cvSVD(MtM,evals,NULL,evects,CV_SVD_MODIFY_A);
#endif
  
  // 最小固有値に対応する固有ベクトルをコピー
  for (int i=0;i<evects->rows;i++)
    cvSetReal1D(ret_vector,i,cvmGet(evects,evects->cols-1,i));
  
  cvReleaseMat(&MtM);
  cvReleaseMat(&evects);
  cvReleaseMat(&evals);
  
  return ret;
}
			
CvMatND *cvCreateTensorEpsilon(int dim){
  // dim 次元の epsilon tensor を作成
  // dim^{dim}
  int *order;
  int i;
  CvMatND *tensor;

  // テンソル作成
  order = (int*)calloc(dim,sizeof(int));
  for (i=0;i<dim;i++)
    order[i] = dim;
  tensor = cvCreateMatND(dim,order,CV_64F);
  
  // 再帰定義により tensor の要素を計算
  for (i=0;i<dim;i++){
    order[0] = i;
    cvCreateMatEpsilonRecur(tensor,dim,1,order);
  }
  
  free(order);

  return tensor;
}

// 再帰定義を用いて episilon tensor を作成
static int cvCreateMatEpsilonRecur(CvMatND *tensor,int dim,int nest,int *order){
  // 最下層の場合
  if (nest == dim){
    cvSetRealND(tensor,order,cvGetEpsilonElement(dim,order));
  }else{
    int i;
    for (i=0;i<dim;i++){
      order[nest] = i;
      cvCreateMatEpsilonRecur(tensor,dim,nest+1,order);
    }
  }
  return 0;
}

static double cvGetEpsilonElement(int dim,int *order){
  // order が偶置換，寄置換かを調べる
  // 単純ソートを利用
  int i,j;
  int rep=0;
  int *copy;

  copy = (int*)calloc(dim,sizeof(int));
  memcpy(copy,order,sizeof(int)*dim);
  
  for (i=0;i<dim;i++){
    if (copy[i] != i){
      for (j=1;j<dim;j++){
	if (copy[j] == i){
	  copy[j] = copy[i];
	  copy[i] = i;
	  rep++;
	  break;
	}
      }
      // 置換が発生していない（order中にiが存在しない）ので0
      if (j==dim)
	return (double)0;
    }
  }
  if ((rep%2)==0){
    // 偶置換
    return 1;
  }
  // それ以外は寄置換
  return -1;
}

int cvPointNormalize(const CvMat *point,
		     CvMat *normalize,
		     CvMat *affine){
  // 平均0,原点からの平均距離が sqrt(2)になるように正規化
  // affine には変換に使用した行列を格納
  // cvmGet(point,i,j) : i-th point の j-th 座標
  int i,j;
  CvMat *ave,*dis;

  // 平均を計算
  ave = cvCreateMat(point->cols-1,1,CV_64F);
  cvmSet(ave,0,0,0);
  cvmSet(ave,1,0,0);

  for (i=0;i<point->rows;i++){
    for (j=0;j<point->cols-1;j++){
      cvmSet(ave,j,0,
	     cvmGet(ave,j,0) + cvmGet(point,i,j));
    }
  }

  for (i=0;i<point->cols-1;i++)
    cvmSet(ave,i,0,cvmGet(ave,i,0)/(double)point->rows);
  
  // 平均からの距離の平均を計算
  dis = cvCreateMat(point->cols-1,1,CV_64F);
  for (i=0;i<point->cols-1;i++)
    cvmSet(dis,i,0,0);

  for (i=0;i<point->rows;i++){
    for (j=0;j<point->cols-1;j++){
      cvmSet(dis,j,0,
	     cvmGet(dis,j,0) + 
	     sqr(cvmGet(point,i,j) - cvmGet(ave,j,0)));
    }
  }

  for (i=0;i<point->cols-1;i++)
    cvmSet(dis,i,0,
	   sqrt(2*point->rows/cvmGet(dis,i,0)));
  
  // affine 変換行列を作成
  if (affine!=NULL){
    cvZero(affine);
    for (i=0;i<point->cols-1;i++){
      cvmSet(affine,i,i,cvmGet(dis,i,0));
      cvmSet(affine,i,point->cols-1,
	     cvmGet(ave,i,0) * cvmGet(dis,i,0) * (-1));
    }
    cvmSet(affine,point->cols-1,point->cols-1,(double)1);
  }

  // 正規化した点を格納
  for (i=0;i<point->rows;i++){
    for (j=0;j<point->cols-1;j++)
      cvmSet(normalize,i,j,
	     (cvmGet(point,i,j) - cvmGet(ave,j,0)) * cvmGet(dis,j,0));
    cvmSet(normalize,i,point->cols-1,(double)1);
  }

  cvReleaseMat(&ave);
  cvReleaseMat(&dis);

  
  return 0;
}

int cvVecNormalizeL2(CvMat *vec){
  // ベクトルをノルム1に正規化
  int i,dim;
  double norm=0;

  // ベクトルであることを確認
  if (vec->rows!=1 && vec->cols!=1)
    return -1;
  
  if (vec->rows!=1)
    dim = vec->rows;
  else
    dim = vec->cols;

  for (i=0;i<dim;i++)
    norm+=sqr(cvGetReal1D(vec,i));
  
  // ノルムの逆数を計算
  norm = 1.0/sqrt(norm);
  
  for (i=0;i<dim;i++)
    cvSetReal1D(vec,i,cvGetReal1D(vec,i) * norm);
  
  return 0;
}
