#include <cv.h>
#include <highgui.h>
#include <stdio.h>

#include "corner.h"
#include "recover.h"
#include "trifocal_tensor.h"
#include "polygon.h"
#include "texture.h"
#include "correspond.h"

// 実行時は
// ./tensortest 入力画像1 入力画像2 入力画像3
// のように指定する

// 途中経過表示の待ち時間
static int Delay = 0;

// 冗長な表示を行うかどうか
static int Verbose = 1;

// 検出する特徴点の最大数
static int PointNum = 500;

// 画像マッチングに用いる閾値(0 - 1)
// 高いほどマッチングし難くなる
static double CorMin = 0.5;

// 対応点の計算を行う場合の閾値
// 2視点：エピポーラ線との距離の閾値
// 3視点：点変換後の点との距離の閾値
static double Length = 2.5;

/* ----- prototype ----- */
static void drawCorrespond(IplImage *imgIn,
			   int num,CvMyCorrespond *correspond,
			   CvMat *inlier);
static void showCorrespond(IplImage *imgIn,const char *windowname,
			   const char *string,
			   int num, CvMyCorrespond *correspond,
			   CvMat *inlier);
/* ----- ----- ----- */

int TFTensorTest(char *input1,char *input2,char *input3){
  // Trifocal テンソルを計算して視点変換を行う
  // 画像および特徴点
  IplImage *imgIn1,*imgIn2,*imgIn3;
  CvPoint2D32f *corner1,*corner2,*corner3;
  int cnum1 = PointNum, cnum2 = PointNum,cnum3 = PointNum;

  // 各画像間のF
  CvMat *F12,*F13,*F23;

  // 対応点組
  CvMyCorrespond *correspond;
  int cornum;

  // テンソル作成に使用
  CvMatND *trif;
  int sizes[3]={3,3,3};
  CvMat *inlier;

  // パッチの描画
  int trinum;
  CvMyTriangle *tri;

  int i;
  
  // まず3枚の画像を読み込む
  imgIn1 = cvLoadImage(input1,CV_LOAD_IMAGE_COLOR);
  corner1 = (CvPoint2D32f*)calloc(cnum1,sizeof(CvPoint2D32f));
  
  imgIn2 = cvLoadImage(input2,CV_LOAD_IMAGE_COLOR);
  corner2 = (CvPoint2D32f*)calloc(cnum2,sizeof(CvPoint2D32f));
  
  imgIn3 = cvLoadImage(input3,CV_LOAD_IMAGE_COLOR);
  corner3 = (CvPoint2D32f*)calloc(cnum2,sizeof(CvPoint2D32f));
  
  // つぎに特徴点位置を検出
  // corner.c
  cvMyFeatureDetection(imgIn1,corner1,&cnum1);
  cvMyFeatureDetection(imgIn2,corner2,&cnum2);
  cvMyFeatureDetection(imgIn3,corner3,&cnum3);
  
  // 対応点組を計算するために各画像間の F 行列を計算
  // correspond.c, fundamental.c
  // 1 -> 2
  F12 = cvCreateMat(3,3,CV_64F);
  searchCorrespond2(imgIn1,corner1,cnum1,
		    imgIn2,corner2,cnum2,
  		    &correspond,&cornum,F12,CorMin,Length);
  if (Verbose){
    showCorrespond(imgIn1,"result","Image1->2",
		   cornum,correspond,NULL);
    cvWaitKey(Delay);
    free(correspond);
  }

  // 1 -> 3
  F13 = cvCreateMat(3,3,CV_64F);
  searchCorrespond2(imgIn1,corner1,cnum1,
		    imgIn3,corner3,cnum3,
		    &correspond,&cornum,F13,CorMin,Length);
  if (Verbose){
    showCorrespond(imgIn1,"result","Image1->3",
		   cornum,correspond,NULL);
    cvWaitKey(Delay);
    free(correspond);
  }
  
  // 2 - > 3 の F 計算
  F23 = cvCreateMat(3,3,CV_64F);
  searchCorrespond2(imgIn2,corner2,cnum2,
		    imgIn3,corner3,cnum3,
		    &correspond,&cornum,F23,CorMin,Length);
  if (Verbose){
    showCorrespond(imgIn2,"result","Image2->3",
		   cornum,correspond,NULL);
    cvWaitKey(Delay);
    free(correspond);
  }

  // これまでに計算した F を利用して3画像間の対応点を探索
  searchCorrespond3(imgIn1,corner1,cnum1,
		    imgIn2,corner2,cnum2,
		    imgIn3,corner3,cnum3,
		    F12,F13,F23,
		    &correspond,
		    &cornum,CorMin,Length);
  if (Verbose){
    showCorrespond(imgIn1,"result","Image1->2,3",
		   cornum,correspond,NULL);
    cvWaitKey(Delay);
  }
  
  // テンソル計算
  // trifocal_tensor.c
  // 適切な対応点を描画
  // 対応点が正しくとれているかどうか確認

  if (cornum<7){
    fprintf(stderr,"A number of corresponding points is too few.\n");
    fprintf(stderr,"Trifocal tensor cannot be estimated.\n");
    exit(0);
  }

  trif = cvCreateMatND(3,sizes,CV_64F);
  inlier = cvCreateMat(cornum,1,CV_32S);
  findTFTensorRansac(correspond,cornum,
		     inlier,trif,100,5);
  
  if (Verbose){
    showCorrespond(imgIn1,"result","Image1->2,3(inlier)",
		   cornum,correspond,inlier);
    cvWaitKey(Delay);
  }
  // ここまでで tensor 計算は終了
  // ここから結果表示

  // パッチ群作成
  // 対応点群から inlier のみを用いてドロネーアルゴリズムにより
  // 三角パッチ群を作成
  // polygon.c
  tri = makeTriangleSetByDelaunay(cornum,inlier,
				  0,correspond,&trinum);
  // パッチ描画
  for (i=0;i<trinum;i++){
    CvPoint2D32f trans[3];
    int j;
    // 点の対応関係と計算されたテンソルを用いて点変換
    for (j=0;j<3;j++)
      transPointByTFTensor2D32f(trif,
				tri[i].corr[j].point[0],
				tri[i].corr[j].point[1],
				&trans[j]);

    // 変換により得られた点にパッチを描画
    // texture.c
    drawTriangleTexture2(imgIn1,
			 tri[i].corr[0].point[0],
			 tri[i].corr[1].point[0],
			 tri[i].corr[2].point[0],
			 imgIn3,
			 trans[0],trans[1],trans[2]);

  }
  
  cvShowImage("result",imgIn3);
  cvWaitKey(0);

  // 後始末
  free(corner1);
  free(corner2);
  free(corner3);
  cvReleaseImage(&imgIn1);
  cvReleaseImage(&imgIn2);
  cvReleaseImage(&imgIn3);
  cvReleaseMat(&F12);
  cvReleaseMat(&F13);
  cvReleaseMat(&F23);
  free(correspond);
  cvReleaseMatND(&trif);
  cvReleaseMat(&inlier);
  free(tri);
  
  return 0;
}

// 処理結果の途中表示
static void showCorrespond(IplImage *imgIn,const char *windowname,
			   const char *string,
			   int num, CvMyCorrespond *correspond,
			   CvMat *inlier){
  IplImage *imgShow;
  static CvFont font;
  static int init = 0;

  // フォント初期化
  if (!init){
    cvInitFont (&font, CV_FONT_HERSHEY_DUPLEX, 1.0, 1.0);
    init=1;
  }
  
  imgShow = cvCloneImage(imgIn);
  cvNamedWindow(windowname);

  drawCorrespond(imgShow,num,correspond,inlier);

  if (string!=NULL){
    cvPutText(imgShow,string,cvPoint(10,50),&font,
	      cvScalar(0,0,0));
  }
  cvShowImage(windowname,imgShow);
  cvReleaseImage(&imgShow);

  return;
}

static void drawCorrespond(IplImage *imgIn,
			   int num,CvMyCorrespond *correspond,
			   CvMat *inlier){
  // 対応点描画
  int i,j;
  for (i=0;i<num;i++){
    // 1点目を円で描画
    if (inlier == NULL || cvGetReal1D(inlier,i)!=0){
      cvCircle(imgIn,
	       cvPoint(correspond[i].point[0].x,
		       correspond[i].point[0].y),
	       5,cvScalar(255,0,0),2);
      for (j=0;j<correspond[i].num;j++)
	cvLine(imgIn,
	       cvPoint(correspond[i].point[0].x,
		       correspond[i].point[0].y),
	       cvPoint(correspond[i].point[j].x,
		       correspond[i].point[j].y),
	       cvScalar(((j>>2)&1)*255,
			((j>>1)&1)*255,
			((j>>0)&1)*255),2);
    }
  }
  return;
}

int main(int argc,char *argv[]){
  if (argc!=4){
    printf("Usage : %s <input1> <input2> <input3>\n",argv[0]);
    exit(0);
  }

  return TFTensorTest(argv[1],argv[2],argv[3]);
}
