// polygonの作成・描画に関する関数群
// ドロネーグラフからパッチ群を作成
// あまりうまいやり方ではない

// opencv/c/delaunay.c より

#include <cv.h>
#include <stdio.h>

#include "polygon.h"
#include "trifocal_tensor.h"

#define LARGE 10000

#define POINT_CORR_THRE 0.1

static  double sqr(double x){
  return x*x;
}

static double disPoint2D32f(CvPoint2D32f p1,CvPoint2D32f p2){
  return sqrt(sqr(p1.x-p2.x)+sqr(p1.y-p2.y));
}

#define NEAREST_THRE 10

static int getNearestCorrespond(int cornum,
			 int useimage,
			 CvMyCorrespond *correspond,
			 CvPoint2D32f point,
			 CvMyCorrespond *ret_cor){
  // point 最近傍（一致する）の点組を返す
  CvMyCorrespond min_cor;
  int i;
  double min = NEAREST_THRE;
  
  for (i=0;i<cornum;i++){
    double dis = disPoint2D32f(point,
			       correspond[i].point[useimage]);
    if (dis<min){
      min_cor = correspond[i];
      min = dis;
    }
  }
  if (min==NEAREST_THRE)
    return -1;
  
  *ret_cor = min_cor;
  return 0;
}
				    
static int makeTriangleFromEdge(CvSubdiv2DEdge edge,
				int cornum, CvMyCorrespond *correspond,
				int useimage,
				CvNextEdgeType next_flag,
				CvMyTriangle *tri_ret){
  CvSubdiv2DPoint *org_pt;
  CvSubdiv2DEdge current;
  int num=0;
  CvMyTriangle tri_tmp;
  
  current = edge;
  do{
    // ポリゴンの最大点数を超過していないか確認
    if (num>=POL_POINT) 
      return -1;
    org_pt = cvSubdiv2DEdgeOrg(current);
    //org_pt の最近傍点をcorrespondから探索
    if (getNearestCorrespond(cornum,useimage,
			     correspond,org_pt->pt,&(tri_tmp.corr[num]))<0){
      // 対応点が存在しない点を含むパッチであれば -1 を返す
      return -1;
    }

    // 次のエッジを取得
    current = cvSubdiv2DGetEdge((CvSubdiv2DEdge)current,next_flag);
    num++;
  }while(current!=edge);

  // 戻り値を代入
  tri_tmp. num = num;
  *tri_ret = tri_tmp;

  return 0;
}

static int cmpPoint2D32f(CvPoint2D32f pt1,CvPoint2D32f pt2){
  // 点が一致しているかどうかを確認
  if (disPoint2D32f(pt1,pt2)<POINT_CORR_THRE)
    return 0;
  return -1;
}

static int cmpTriangle(CvMyTriangle tri1,CvMyTriangle tri2,int useimage){
  // 三角パッチを比較
  // 一致すれば 0 を返す
  unsigned char used[POL_POINT];
  int ret=0;
  int i,j;
  
  // 点数
  if (tri1.num!=tri2.num) return -1;

  for (i=0;i<tri1.num;i++)
    used[i] = 0;
  
  // 一致する対応点を取得
  for (i=0;i<tri1.num;i++)
    for(j=0;j<tri2.num;j++)
      if (!used[j] && !cmpPoint2D32f(tri1.corr[i].point[useimage],
				     tri2.corr[j].point[useimage])){
	  ret++;
	  used[j] = 1;
      }
  if (ret==tri1.num)
    return 0;

  return -1;
}
      

static CvMyTriangle *elimDuplicateTriangle(int tri_num,
					   CvMyTriangle *tri_set,
					   int *ret_num){
  // 重複の三角パッチを削除
  int *dupl,dupnum=0;
  int new_num=0;
  int i,j;
  CvMyTriangle *tri_ret;

  if (tri_num==0)
    return NULL;

  dupl = (int*)calloc(tri_num,sizeof(int));

  for (i=0;i<tri_num-1;i++)
    if (dupl[i]==0)
      for (j=i+1;j<tri_num;j++)
	if (dupl[j]!=1 && !cmpTriangle(tri_set[i],tri_set[j],0)){
	  dupl[j] = 1;
	  dupnum++;
	}
  
  // 必用分だけメモリ確保しなおし
  tri_ret = (CvMyTriangle*)calloc(tri_num-dupnum,sizeof(CvMyTriangle));
  new_num=0;
  for (i=0;i<tri_num;i++){
    if (dupl[i]==0){
      tri_ret[new_num] = tri_set[i];
      new_num++;
    }
  }

  free(dupl);
  *ret_num = new_num;
  
  return tri_ret;
}

static CvMyTriangle *makeTriangleSetFromGraph(CvSubdiv2D *graph,
					      int cornum,
					      int useimage,
					      CvMyCorrespond *correspond,
					      int *tri_num_ret){
  
  CvSeqReader  reader;
  CvMyTriangle *tri_set,tri_tmp,*tri_ret;
  int tri_num = 0;
  
  int i, total = graph->edges->total;
  int elem_size = graph->edges->elem_size;
  
  
  // 戻り値用の領域を最大限確保
  tri_set = (CvMyTriangle *)calloc(total*2,sizeof(CvMyTriangle));
  
  cvStartReadSeq( (CvSeq*)(graph->edges), &reader, 0 );
    
  for( i = 0; i < total; i++ ){
    CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);
    if( CV_IS_SET_ELEM( edge )){
      // 左側のパッチを作成
      if (makeTriangleFromEdge((CvSubdiv2DEdge)edge,cornum,correspond,
			       useimage,CV_NEXT_AROUND_LEFT,&tri_tmp)==0){
	tri_set[tri_num] = tri_tmp;
	tri_num++;
      }
      
      // 右側のパッチを作成
      if (makeTriangleFromEdge((CvSubdiv2DEdge)edge,cornum,correspond,
			       useimage,CV_NEXT_AROUND_RIGHT,&tri_tmp)==0){
	tri_set[tri_num] = tri_tmp;
	tri_num++;
      }
    }
    CV_NEXT_SEQ_ELEM( elem_size, reader );
  }

  // 重複パッチを削除
  tri_ret = elimDuplicateTriangle(tri_num,tri_set,tri_num_ret);

  free(tri_set);
  
  return tri_ret;
}

CvMyTriangle *makeTriangleSetByDelaunay(int cornum,CvMat *inlier,
					int useimage,
					CvMyCorrespond *correspond,
					int *trinum){
  // 対応点組から三角パッチ群を作成
  // ドロネー網に関する関数を使用

  // cornum : 対応点の組数
  // usepoint : 使用する画像の番号
  // correspond : 対応点の集合

  // グラフの作成に利用
  CvSubdiv2D *graph;
  CvMemStorage *storage = cvCreateMemStorage(0);
  
  // 矩形領域の決定に利用
  int minx=LARGE,miny=LARGE,maxx=-LARGE,maxy=-LARGE;

  CvMyTriangle *ret_tri;

  int i;

  if (cornum<=0) return NULL;
  if (useimage<0 || correspond[0].num<=useimage) return NULL;
  
  // 最大矩形を作成
  for (i=0;i<cornum;i++){
    if (minx > correspond[i].point[useimage].x)
      minx = correspond[i].point[useimage].x;
    if (maxx < correspond[i].point[useimage].x)
      maxx = correspond[i].point[useimage].x;
    if (miny > correspond[i].point[useimage].y)
      miny = correspond[i].point[useimage].y;
    if (maxy < correspond[i].point[useimage].y)
      maxy = correspond[i].point[useimage].y;
  }

  
  CvRect rect = cvRect(minx-100,miny-100,
		       maxx-minx+200,maxy-miny+200);
  graph = cvCreateSubdivDelaunay2D(rect,storage);

  // ドロネー網作成
  for (i=0;i<cornum;i++){
    if (inlier==NULL || cvGetReal1D(inlier,i))
      cvSubdivDelaunay2DInsert(graph,correspond[i].point[useimage]);
  }

  // ドロネー網からパッチ群を作成
  ret_tri = makeTriangleSetFromGraph(graph,cornum,useimage,correspond,trinum);
  
  // グラフ関連を解放
  cvReleaseMemStorage(&storage);

  return ret_tri;
}
