/* ===== Trifocal tensor 計算 ===== */
#include <cv.h>
#include <stdio.h>

#include "trifocal_tensor.h"
#include "linear.h"

#define USE_POINT_NUM 7

static  double sqr(double x){
  return x*x;
}

static  CvMat *cvCreateSimMat(const CvMat *mat){
  return cvCreateMat(mat->rows,mat->cols,mat->type);
}

/* ----- prototype ----- */
static int cvMatShow(const CvMat *mat);
static int random_swap(int num,int *list);
static int cvCopyRow(const CvMat *src,int srow,
		     CvMat *dst,int drow);
static int cvNormalizeHomo(CvMat *src);
static int cvmSetPointRow(CvMat *mat,int row,CvPoint2D32f pt);
/* ----- ----- ----- */


int findTFTensorRansac(CvMyCorrespond *correspond, int num,
		       CvMat *inlier_ret,
		       CvMatND *tftensor,
		       int loopnum,double lenght){
  // 対応点組から RANSAC を利用して trifocal tensor を計算
  // correspond : 対応点組
  // num : 対応点組の数
  // inlier_ret : RANSACにより得られる inlier のフラグを格納
  // tftensor : 戻り値となる trifocal tensor
  // loopnum : ransac のループ回数を指定，多いとロバスト性が向上
  // length : 点変換後の点との距離の閾値
  CvMat *mat[3];
  int *inlier;
  int i,j;

  // 対応点群を行列形式に変換
  for (i=0;i<3;i++)
    mat[i] = cvCreateMat(num,3,CV_64F);

  for (i=0;i<num;i++)
    for (j=0;j<3;j++)
      cvmSetPointRow(mat[j],i,correspond[i].point[j]);

  inlier = (int*)calloc(num,sizeof(int));
  // tensor 計算
  cvFindTFTensorRANSAC(mat[0],mat[1],mat[2],num,inlier,
		       tftensor,loopnum,lenght);
  // inlier を変換
  if (inlier!=NULL)
    for (i=0;i<num;i++)
      cvSetReal1D(inlier_ret,i,inlier[i]);

  //後始末
  free(inlier);
  for (i=0;i<3;i++)
    cvReleaseMat(&mat[i]);
  
  return 0;
}

CvMatND *cvFindTFTensorLinear2(const CvMat *point1,
				     const CvMat *point2,
				     const CvMat *point3){
  // 線形演算を利用した３焦点テンソルの計算
  // Multiple View Geometry Chap.16 より
  // point1->rows == poinr2->rows == point3->rows
  int i,j,k,q,r,s,t;
  int a,b,c;
  CvMat *M,*solut;
  CvMatND *eps;
  CvMatND *tensor;
  int sizes[3]={3,3,3};

  eps = cvCreateTensorEpsilon(3);
  tensor = cvCreateMatND(3,sizes,CV_64F);

  M = cvCreateMat(point1->rows * 9,27,CV_64F);

  // 点の数に関するループ(aで指定)
  for (a=0;a<point1->rows;a++){
    // 1つの対応点組より得られる式のループ（9式,bで指定）
    for (s=0,b=0;s<3;s++){
      for (t=0;t<3;t++,b++){
	// 未知数に関するループ(27変数，cで指定)
	for (c=0,q=0;q<3;q++){
	  for (r=0;r<3;r++){
	    for (i=0;i<3;i++,c++){
	      // 既知数に関するループ
	      double sum=0;
	      for (j=0;j<3;j++){
		for (k=0;k<3;k++){
		  sum += 
		    cvGetReal2D(point1,a,i) *
		    cvGetReal2D(point2,a,j) * 
		    cvGetReal2D(point3,a,k) *
		    cvGetReal3D(eps,j,q,s) *
		    cvGetReal3D(eps,k,r,t);
		}
	      }
	      cvmSet(M,a*9+b,c,sum);
	    }
	  }
	}
      }
    }
  }
  
  // 最小固有値に対応する固有ベクトルを計算
  solut = cvCreateMat(27,1,CV_64F);
  cvGetMinEigenVector(M,solut);
  
  // tensor に格納
  for (c=0,q=0;q<3;q++)
    for (r=0;r<3;r++)
      for (i=0;i<3;i++,c++){
	cvSetReal3D(tensor,q,r,i,cvmGet(solut,c,0));
      }
  
  // M 解放
  cvReleaseMat(&M);
  cvReleaseMat(&solut);
  cvReleaseMatND(&eps);
  
  return tensor;
}

int transPointByTFTensor2D32f(const CvMatND *tensor,
			      const CvPoint2D32f pt1,
			      const CvPoint2D32f pt2,
			      CvPoint2D32f *pt3){
  // trifocal tensor を利用して点変換
  // tensor : trifocal tensor
  // pt1, pt2 : 画像1,2の特徴点
  // pt3 : 変換後の点が格納
  CvMat *point1,*point2,*point3;

  point1 = cvCreateMat(3,1,CV_64F);
  point2 = cvCreateMat(3,1,CV_64F);
  point3 = cvCreateMat(3,1,CV_64F);
  
  cvmSet(point1,0,0,pt1.x);
  cvmSet(point1,1,0,pt1.y);
  cvmSet(point1,2,0,(double)1);
  
  cvmSet(point2,0,0,pt2.x);
  cvmSet(point2,1,0,pt2.y);
  cvmSet(point2,2,0,(double)1);

  cvTFTensorPointTransfer(tensor,point1,point2,point3);

  pt3->x = cvmGet(point3,0,0)/cvmGet(point3,2,0);
  pt3->y = cvmGet(point3,1,0)/cvmGet(point3,2,0);

  return 0;
}

int cvTFTensorPointTransfer(const CvMatND *tensor,
			    const CvMat *point1,
			    const CvMat *point2,
			    CvMat *point3){
  //  trifocal tensor を用いた点変換
  // point3 なので，変数 k についてまとめる
  // p. 391 より
  int i,j,k,q,r,s,t;
  int a;
  CvMat *M;
  CvMatND *eps;

  eps = cvCreateTensorEpsilon(3);
  
  M = cvCreateMat(9,3,CV_64F);
  // 1つの点組より得られる式に関するループ(9式，aで指定)
  for (a=0,s=0;s<3;s++)
    for (t=0;t<3;t++,a++){
      // 未知変数に対するループ(3変数)
      for (k=0;k<3;k++){
	// 既知数に関するループ
	double val = 0;
	for (i=0;i<3;i++)
	  for (j=0;j<3;j++)
	    for (q=0;q<3;q++)
	      for (r=0;r<3;r++){
		val += 
		  cvGetReal1D(point1,i) *
		  cvGetReal1D(point2,j) *
		  cvGetReal3D(eps,j,q,s) *
		  cvGetReal3D(eps,k,r,t) *
		  cvGetReal3D(tensor,q,r,i);
	      }
	cvmSet(M,a,k,val);
      }
    }
  cvGetMinEigenVector(M,point3);

  cvReleaseMat(&M);
  cvReleaseMatND(&eps);
    
  return 0;
}

int cvTFTensorFindEpipoles(const CvMatND *tensor,
		     CvMat *epi1,
		     CvMat *epi2){
  // エピポールを計算
  // p. 394 より
  // epi1 = e'
  // epi2 = e''
  int i,q,r;
  CvMat *T,*Tt,*v,*V1,*V2;
  
  T = cvCreateMat(3,3,CV_64F);
  Tt = cvCreateMat(3,3,CV_64F);

  V1 = cvCreateMat(3,3,CV_64F);
  V2 = cvCreateMat(3,3,CV_64F);
  v = cvCreateMat(3,1,CV_64F);
  
  for (i=0;i<3;i++){
    for (q=0;q<3;q++)
      for (r=0;r<3;r++){
	cvmSet(T,q,r,cvGetReal3D(tensor,q,r,i));
	cvmSet(Tt,r,q,cvGetReal3D(tensor,q,r,i));
      }
    
    cvGetMinEigenVector(T,v);
    for (r=0;r<3;r++)
      cvmSet(V2,i,r,cvmGet(v,r,0));
    
    cvGetMinEigenVector(Tt,v);
    for (q=0;q<3;q++)
      cvmSet(V1,i,q,cvmGet(v,q,0));
  }
  
  // V1, V2 の左零空間よりエピポールを計算
  cvGetMinEigenVector(V1,epi1);
  cvGetMinEigenVector(V2,epi2);
  
  // 解放
  cvReleaseMat(&T);
  cvReleaseMat(&Tt);
  cvReleaseMat(&V1);
  cvReleaseMat(&V2);
  cvReleaseMat(&v);
  
  return 0;
}

int cvTFTensorFindCameraMat(const CvMatND *tensor,
			     CvMat *cam1,
			     CvMat *cam2){
  // trifocal tensor よりカメラ行列を作成
  CvMat *epi1,*epi2;
  CvMat *E1,*E2,*E1E2;
  int i,j,q,r;
  
  // まずエピポールを計算
  epi1 = cvCreateMat(3,1,CV_64F);
  epi2 = cvCreateMat(3,1,CV_64F);
  
  cvTFTensorFindEpipoles(tensor,epi1,epi2);

  // epi1, epi2 をノルム1に正規化
  cvVecNormalizeL2(epi1);
  cvVecNormalizeL2(epi2);
  
  // epipole からカメラ行列を計算
  // P' = [[T1,T2,T3]e''|e']
  for (i=0;i<3;i++){
    for (q=0;q<3;q++){
      double sum = 0;
      for (r=0;r<3;r++)
	sum += cvGetReal3D(tensor,q,r,i) * cvmGet(epi2,r,0);
      cvmSet(cam1,q,i,sum);
    }
    cvmSet(cam1,i,3,cvmGet(epi1,i,0));
  }
  
  // 3台目のカメラを計算
  // P'' = [(e''e''^T - I)[T1^T,T2^T,T3^T]e'|e'']
  //     = [E1E2e'|e'']
  E1 = cvCreateMat(3,3,CV_64F);
  for (i=0;i<3;i++)
    for (j=0;j<3;j++){
      cvmSet(E1,i,j,cvmGet(epi2,i,0) * cvmGet(epi2,j,0));
      if (i==j)
	cvmSet(E1,i,j,cvmGet(E1,i,j)-(double)1);
    }
  
  E2 = cvCreateMat(3,3,CV_64F);
  for (i=0;i<3;i++){
    for (r=0;r<3;r++){
      double sum = 0;
      for (q=0;q<3;q++)
	sum += cvGetReal3D(tensor,q,r,i) * cvmGet(epi1,q,0);
      cvmSet(E2,r,i,sum);
    }
  }

  E1E2 = cvCreateMat(3,3,CV_64F);
  cvMatMul(E1,E2,E1E2);
  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      cvmSet(cam2,i,j,cvmGet(E1E2,i,j));
  for (i=0;i<3;i++)
    cvmSet(cam2,i,3,cvmGet(epi2,i,0));
  
  cvReleaseMat(&E1);
  cvReleaseMat(&E2);
  cvReleaseMat(&epi1);
  cvReleaseMat(&epi2);
  
  return 0;
}

int cvFindTFTensorNormalize(const CvMat *point1,
			    const CvMat *point2,
			    const CvMat *point3,
			    CvMatND *tensor){
  // 点の正規化を行った後 tensor を計算する
  const CvMat *pset[3];
  CvMat *normalize[3];
  CvMat *affine[3],*inv[3];
  CvMatND *norm_tensor;
  int i,j,k,r,s,t;
  
  // 正規化された点を計算
  pset[0] = point1;
  pset[1] = point2;
  pset[2] = point3;

  for (i=0;i<3;i++){
    normalize[i] = cvCreateSimMat(pset[i]);
    affine[i] = cvCreateMat(3,3,CV_64F);
    cvPointNormalize(pset[i],normalize[i],affine[i]);
    inv[i] = cvCreateSimMat(affine[i]);
    cvInvert(affine[i],inv[i],CV_SVD);
  }

  // 正規化された点をもとに Trifocal tensor を計算
  norm_tensor = cvFindTFTensorLinear2(normalize[0],
				      normalize[1],
				      normalize[2]);
  
  // 得られたテンソルを元の空間に戻す
  // T^jk_i = H^r_i H'^-1^j_s H''^-1^k_t T^st_r
  // 式に関するループ(27式)
  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      for (k=0;k<3;k++){
	// 既知数に関するループ
	double sum = 0;
	for (r=0;r<3;r++)
	  for (s=0;s<3;s++)
	    for (t=0;t<3;t++){
	      sum += 
		cvmGet(affine[0],r,i) *
		cvmGet(inv[1],j,s) *
		cvmGet(inv[2],k,t) *
		cvGetReal3D(norm_tensor,s,t,r);
	    }
	cvSetReal3D(tensor,j,k,i,sum);
      }
  
  // 解放
  for (i=0;i<3;i++){
    cvReleaseMat(&normalize[i]);
    cvReleaseMat(&affine[i]);
    cvReleaseMat(&inv[i]);
  }
  
  cvReleaseMatND(&norm_tensor);

  return 0;
}

int evalTFTensor(const CvMat *point1,
		 const CvMat *point2,
		 const CvMat *point3,
		 int num,int *inliers,
		 CvMatND *tensor,double thre){
  // tensor の評価
  // 1,2 -> 3への点変換を利用して評価
  CvMat *src1 = cvCreateMat(1,3,CV_64F);
  CvMat *src2 = cvCreateMat(1,3,CV_64F);
  CvMat *trans = cvCreateMat(1,3,CV_64F);
  int support = 0;
  double dis;
  int i;

  for (i=0;i<num;i++){
    cvCopyRow(point1,i,src1,0);
    cvCopyRow(point2,i,src2,0);
    cvTFTensorPointTransfer(tensor,src1,src2,trans);
    cvNormalizeHomo(trans);
    cvmSet(trans,0,0,cvmGet(trans,0,0)/cvmGet(trans,0,2));
    cvmSet(trans,0,1,cvmGet(trans,0,1)/cvmGet(trans,0,2));
    cvmSet(trans,0,2,1);
    dis =
      sqrt(sqr(cvmGet(point3,i,0) - cvmGet(trans,0,0)) +
	   sqr(cvmGet(point3,i,1) - cvmGet(trans,0,1)));
    if (inliers!=NULL)
      inliers[i] = 0;
    if (dis<thre){
      support++;
      if (inliers!=NULL)
	inliers[i] = 1;
      
    }
  }

  cvReleaseMat(&src1);
  cvReleaseMat(&src2);
  cvReleaseMat(&trans);
  
  return support;
}

int cvFindTFTensorRANSAC(const CvMat *point1,
			 const CvMat *point2,
			 const CvMat *point3,
			 int num,
			 int *inliers,
			 CvMatND *tensor,
			 int maxloop,double length){
  // RANSAC を用いて TFTenosr を計算
  int *order;
  int i,j,loop;
  int maxsup=0;
  const CvMat *pset[3];
  CvMat *est[3];
  CvMatND *temp_tens;
  int size[3]={3,3,3};
  CvMat *trans = cvCreateMat(1,3,CV_64F);
  int support;

  // pset に格納
  pset[0] = point1;
  pset[1] = point2;
  pset[2] = point3;
  
  // 推定に使用する点列を作成
  for (i=0;i<3;i++)
    est[i] = cvCreateMat(USE_POINT_NUM,3,CV_64F);
  
  // 使用する点のリストを作成
  order = (int*)calloc(num,sizeof(int));
  for (i=0;i<num;i++)
    order[i] = i;
  
  // 仮のテンソルを格納する領域確保
  temp_tens = cvCreateMatND(3,size,CV_64F);

  for (loop=0;loop<maxloop;loop++){
    // リストをランダムに入れ替え(使用する点を決定)
    random_swap(num,order);

    // 使用する点を格納
    for (i=0;i<3;i++)
      for (j=0;j<est[i]->rows;j++){
	cvCopyRow(pset[i],order[j],est[i],j);
      }
    // est を使用してTFtensor 計算
    cvFindTFTensorNormalize(est[0],est[1],est[2],temp_tens);
    
    // テンソルを評価(1,2 -> 3 の点変換を用いて評価)
    support = evalTFTensor(point1,point2,point3,num,inliers,
			   temp_tens,length);
    if (support>maxsup){
      cvCopy(temp_tens,tensor);
      maxsup = support;
    }
  }

  // 最終テンソルを評価(1,2 -> 3 の点変換を用いて評価)
  // inlier を決定
  support = evalTFTensor(point1,point2,point3,num,inliers,
			 tensor,length);
  // 後始末
  for (i=0;i<3;i++)
    cvReleaseMat(&est[i]);
  cvReleaseMat(&trans);
  cvReleaseMatND(&temp_tens);
  free(order);
		 
  return 0;
}
			 
static int cvMatShow(const CvMat *mat){
  int i,j;

  for (i=0;i<mat->rows;i++){
    for (j=0;j<mat->cols;j++)
      printf("%f ",cvmGet(mat,i,j));
    printf("\n");
  }
  
  return 0;
}

static int random_swap(int num,int *list){
  // list の中身をランダムに入れ替え
  int i,r;
  int swap;
  static CvRNG rngstate = cvRNG(0xffffffff);

  for (i=0;i<num;i++){
    r = (int)(cvRandInt(&rngstate) / (double) 0xffffffff * num);
    swap = list[i];
    list[i] = list[r];
    list[r] = swap;
  }
  
  return 0;
}


static int cvCopyRow(const CvMat *src,int srow,
		     CvMat *dst,int drow){
  int i;
  
  for (i=0;i<src->cols;i++)
    cvmSet(dst,drow,i,cvmGet(src,srow,i));
  return 0;
}

static int cvNormalizeHomo(CvMat *src){
  int i,j;

  for (i=0;i<src->rows;i++){
    for (j=0;j<src->cols-1;j++)
      cvmSet(src,i,j,cvmGet(src,i,j)/cvmGet(src,i,src->cols-1));
    cvmSet(src,i,src->cols-1,1);
  }
  return 0;
}

int cvMatNDShow(CvMatND *tensor){
  int i,j,k;

  for (i=0;i<3;i++){
    for (j=0;j<3;j++){
      for (k=0;k<3;k++)
	printf("%e ",cvGetReal3D(tensor,i,j,k));
      printf("\n");
    }
    printf("\n");
  }
  return 0;
}

static int cvmSetPointRow(CvMat *mat,int row,CvPoint2D32f pt){
  if (mat==NULL || mat->rows<=row) return -1;
  cvmSet(mat,row,0,pt.x);
  cvmSet(mat,row,1,pt.y);

  if (mat->cols==3)
    cvmSet(mat,row,2,1);
  
  return 0;
}

